@@ -2,6 +2,7 @@
 #
 # SPDX-License-Identifier: EUPL-1.2
 
+import statsd
 from .. import Base
 from ..infrastructure import InfrastructureFactory
 from ..repository import RepositoryFactory
@@ -9,10 +10,25 @@ from .events import Event, EventService, event
 from abc import ABC, abstractmethod
 from copy import copy
 from functools import partial
-from typing import Dict
+from typing import Any, Dict, Set, Type, cast
 from uuid import UUID
 
-__all__ = ["event", "Event", "EventService", "UserInfo"]
+__all__ = [
+    "build_command_lookup_table",
+    "CommandBase",
+    "CQRS",
+    "DomainCommandContainer",
+    "DomainQueryContainer",
+    "event",
+    "Event",
+    "EventService",
+    "MiddlewareBase",
+    "SplitCommandBase",
+    "SplitQueryBase",
+    "QueryBase",
+    "QueryMiddlewareBase",
+    "UserInfo",
+]
 
 
 class UserInfo:
@@ -29,6 +45,214 @@ class UserInfo:
         self.permissions = permissions
 
 
+class SplitCommandBase(Base, ABC):
+    """
+    Base class for 'command' parts of domains.
+
+    This version should be used when every command is split out to its own class.
+    """
+
+    @property
+    @abstractmethod
+    def name(self) -> str:
+        raise NotImplementedError
+
+    @abstractmethod
+    def __call__(self) -> None:
+        raise NotImplementedError
+
+    @property
+    def cmd(self) -> Any:
+        """
+        Reference to the DomainCommandContainer instance that this command is part of.
+
+        Can be used to call other commands.
+        """
+        return self._command_instance
+
+    @cmd.setter
+    def cmd(self, value: "DomainCommandContainer"):
+        self._command_instance = value
+
+    def __init__(
+        self,
+        repository_factory: RepositoryFactory,
+        context: str,
+        user_uuid: UUID,
+        event_service: EventService,
+        command_instance: "DomainCommandContainer",
+    ):
+        self.repository_factory = repository_factory
+        self.context = context
+        self.user_uuid = user_uuid
+        self.event_service = event_service
+        self.cmd = command_instance
+
+    def get_repository(self, name: str):
+        """Initialize and retrieve a repository by name.
+
+        This also passes the event service to the repository constructor.
+
+        :param name: Name of the repository to retrieve
+        :return: Initialized repository
+        """
+
+        return self.repository_factory.get_repository(
+            name=name, context=self.context, event_service=self.event_service
+        )
+
+
+class SplitQueryBase(Base, ABC):
+    """
+    Base class for 'query' parts of domains.
+
+    This version should be used when every query is split out to its own class.
+    """
+
+    @property
+    @abstractmethod
+    def name(self) -> str:
+        raise NotImplementedError
+
+    @abstractmethod
+    def __call__(self) -> Any:
+        raise NotImplementedError
+
+    @property
+    def qry(self) -> Any:
+        """
+        Reference to the DomainQueryContainer instance that this query is part of.
+
+        Can be used to call other queries.
+        """
+        return self._query_instance
+
+    @qry.setter
+    def qry(self, value: "DomainQueryContainer"):
+        self._query_instance = value
+
+    def __init__(
+        self,
+        repository_factory: RepositoryFactory,
+        context: str,
+        user_uuid: UUID,
+        query_instance: "DomainQueryContainer",
+    ):
+        self.repository_factory = repository_factory
+        self.context = context
+        self.user_uuid = user_uuid
+        self.qry = query_instance
+
+    def get_repository(self, name: str):
+        """Initialize and retrieve a repository by name.
+
+        This does not pass an event service to the repository constructor, so
+        no events can be triggered accidentally.
+
+        :param name: Name of the repository to retrieve
+        :return: Initialized repository
+        """
+
+        return self.repository_factory.get_repository(
+            name=name, context=self.context, event_service=None
+        )
+
+
+_CommandLookupTable = Dict[str, Type[SplitCommandBase]]
+_QueryLookupTable = Dict[str, Type[SplitQueryBase]]
+
+
+def build_command_lookup_table(
+    commands: Set[Type[SplitCommandBase]],
+) -> _CommandLookupTable:
+    """
+    Build a command lookup table (keyed on the command's "short name") containing
+    the listed commands.
+    """
+
+    command_lookup_table: _CommandLookupTable = {}
+    for command in commands:
+        assert (
+            command.name not in command_lookup_table
+        ), f"Tried to register second command with name={command.name}"
+
+        command_lookup_table[getattr(command, "name")] = command
+
+    return command_lookup_table
+
+
+def build_query_lookup_table(
+    queries: Set[Type[SplitQueryBase]],
+) -> _QueryLookupTable:
+    """
+    Build a lookup table for queries (keyed on the query's "short name")
+    containing the listed queries.
+    """
+
+    query_lookup_table: _QueryLookupTable = {}
+    for query in queries:
+        assert (
+            query.name not in query_lookup_table
+        ), f"Tried to register second query with name={query.name}"
+
+        query_lookup_table[getattr(query, "name")] = query
+
+    return query_lookup_table
+
+
+class DomainCommandContainer:
+    def __init__(
+        self,
+        repository_factory: RepositoryFactory,
+        context: str,
+        user_uuid: UUID,
+        event_service: EventService,
+        command_lookup_table: _CommandLookupTable,
+    ):
+        self.repository_factory = repository_factory
+        self.context = context
+        self.user_uuid = user_uuid
+        self.event_service = event_service
+        self.command_lookup_table = command_lookup_table
+
+    def __getattr__(self, name: str) -> SplitCommandBase:
+        try:
+            return self.command_lookup_table[name](
+                repository_factory=self.repository_factory,
+                context=self.context,
+                user_uuid=self.user_uuid,
+                event_service=self.event_service,
+                command_instance=self,
+            )
+        except KeyError as e:
+            raise AttributeError(name) from e
+
+
+class DomainQueryContainer:
+    def __init__(
+        self,
+        repository_factory: RepositoryFactory,
+        context: str,
+        user_uuid: UUID,
+        query_lookup_table: _QueryLookupTable,
+    ):
+        self.repository_factory = repository_factory
+        self.context = context
+        self.user_uuid = user_uuid
+        self.query_lookup_table = query_lookup_table
+
+    def __getattr__(self, name: str) -> SplitQueryBase:
+        try:
+            return self.query_lookup_table[name](
+                repository_factory=self.repository_factory,
+                context=self.context,
+                user_uuid=self.user_uuid,
+                query_instance=self,
+            )
+        except KeyError as e:
+            raise AttributeError from e
+
+
 class CommandBase(Base, ABC):
     "Base class for 'command' parts of domains"
 
@@ -247,17 +471,21 @@ class QueryWrapper(Base):
         :return: wrapped method
         :rtype: callable
         """
-        original_attribute = self.query_instance.__getattribute__(attr)
+        original_attribute = getattr(self.query_instance, attr)
+
         if callable(original_attribute):
 
             def wrapped(*args, **kwargs):
                 """Return attribute wrapped in middlewares."""
 
-                with self.statsd.get_timer("TotalQueryTime").time(attr):
+                timer = cast(
+                    statsd.Timer, self.statsd.get_timer("TotalQueryTime")
+                )
+                with timer.time(attr):
                     wrapped_func = partial(original_attribute, *args, **kwargs)
 
                     for middleware in self.middleware:
-                        wrapped_func = partial(middleware, func=wrapped_func)
+                        wrapped_func = partial(middleware, wrapped_func)
                     return wrapped_func()
 
             return wrapped
@@ -295,16 +523,18 @@ class CommandWrapper(Base):
         :rtype: callable
         """
 
-        original_attribute = self.command_instance.__getattribute__(attr)
+        original_attribute = getattr(self.command_instance, attr)
         if callable(original_attribute):
 
             def wrapped(*args, **kwargs):
                 """Return attribute wrapped in middlewares."""
-                with self.statsd.get_timer("TotalCommandTime").time(attr):
-
+                timer = cast(
+                    statsd.Timer, self.statsd.get_timer("TotalCommandTime")
+                )
+                with timer.time(attr):
                     wrapped_func = partial(original_attribute, **kwargs)
                     for middleware in self.middlewares:
-                        wrapped_func = partial(middleware, func=wrapped_func)
+                        wrapped_func = partial(middleware, wrapped_func)
 
                     wrapped_func()
                 return
@@ -417,7 +647,8 @@ class CQRS(Base):
             + f"'{context}' for user '{user_uuid}'"
         )
 
-        with self.statsd.get_timer(domain).time("get_query_instance"):
+        timer = cast(statsd.Timer, self.statsd.get_timer(domain))
+        with timer.time("get_query_instance"):
             query_instance = self.domains[domain]["module"].get_query_instance(
                 self.domains[domain]["repository_factory_ro"],
                 context=context,
@@ -468,7 +699,8 @@ class CQRS(Base):
             + f"'{context}' for user '{user_uuid}'"
         )
 
-        with self.statsd.get_timer(domain).time("get_command_instance"):
+        timer = cast(statsd.Timer, self.statsd.get_timer(domain))
+        with timer.time("get_command_instance"):
             event_service = EventService(
                 correlation_id=correlation_id,
                 domain=domain,
