@@ -28,7 +28,7 @@ class RepositoryBase(Base):
 
 
 class Repository(RepositoryBase):
-    def save(self):
+    def save(self, user_info=None):
         """ Uses the mapping _events_to_calls for calling the repo methods"""
 
         for ev in self.event_service.get_events_by_type(
@@ -38,12 +38,11 @@ class Repository(RepositoryBase):
                 self, self._events_to_calls[ev.event_name]
             )
 
-            method_to_call(ev)
+            method_to_call(ev, user_info)
 
 
 class RepositoryFactory(Base):
-    """Create context-specific "repository" instances for domains
-    """
+    """Create context-specific "repository" instances for domains"""
 
     __slots__ = ["infrastructure_factory", "repositories"]
 
